@@ -1,6 +1,6 @@
-import os
 import re
 import subprocess
+from importlib import resources
 
 import yaml
 
@@ -8,68 +8,98 @@ from . import config
 
 
 class AppMan:
-    def __init__(self, path, config=config.CONFIG_PATH):
-        self.path = path
-        self.config = config
-        self.packages = []
+    def __init__(self):
+        self.config = None
         self.formulas = []
-        self.load_data(path)
+        self.packages = []
+        self.user_packages = []
+        self.load_data()
+
+    def load_data(self):
 
-    def load_data(self, path):
         # config
-        cfdata, _ = self._load_data_file(self.config)
+        cfdata = self._load_data_resource(config.DATA_PKG, config.CONFIG_RES_YAML)
         self.config = Config(cfdata)
 
-        # package managers
-        for pmdata, pmpath in list(self._load_data_dir(config.PM_DIR)):
-            pmname = self._get_name_from_path(pmpath)
-            pm = Formula(pmname)
-            pm.load(pmdata)
-            self.formulas.append(pm)
-
         # formulas
-        for fdata, fpath in self._load_data_dir(config.FORMULAS_DIR):
-            fname = self._get_name_from_path(fpath)
-            formula = Formula(fname, custom=True)
-            formula.load(fdata)
+        for ffile in self._get_data_resource_files(config.FORMULAS_PKG):
+            data = self._load_data_resource(config.FORMULAS_PKG, ffile.name)
+            custom = data["type"] == "custom"
+            formula = Formula(ffile.stem, custom=custom)
+            formula.load(self._load_data_resource(config.FORMULAS_PKG, ffile.name))
             self.formulas.append(formula)
 
-        # packages: cli
-        for package in self._create_packages(path, "cli"):
-            self.packages.append(package)
+        # packages
+        for pt in config.PACKAGES_TYPES:
+            pkg = f"{config.PACKAGES_PKG}.{pt['pkg']}"
+            for pfile in self._get_data_resource_files(pkg):
+                data = self._load_data_resource(pkg, pfile.name)
+                package = Package(pfile.stem, pt["id"])
+                package.load(data)
+                self.packages.append(package)
 
-        # packages: gui
-        for package in self._create_packages(path, "gui"):
-            self.packages.append(package)
-
-        # packages: group by files
-        for presult in self._load_data_grouped(path):
-            package = Package(presult["id"], presult["ptype"])
+        # user packages
+        for presult in self._get_grouped_data_resource_files(config.USER_PKG):
+            package = UserPackage(presult["id"], presult["ptype"])
             package.load(presult["data"])
-            self.packages.append(package)
+            self.user_packages.append(package)
 
-    def get_packages(self, os, package_type, label=None):
-        return [
-            package
-            for package in self.packages
+    def add_user_package(self, package, labels=None):
+        # add default labels for package
+        if labels:
+            labels.extend(package.labels)
+        else:
+            labels = package.labels
+
+        user_package = UserPackage(package.id, package.type, labels)
+        resource = self._get_resource_name(package.type)
+        self._add_data_resource(config.USER_PKG, resource, user_package.data)
+
+    def delete_user_package(self, user_package):
+        resource = self._get_resource_name(user_package.type)
+        self._delete_data_resource(config.USER_PKG, resource, user_package.data)
+
+    def get_user_packages(self, package_type, id=None, labels=None):
+        packages = []
+        for package in self.user_packages:
             if (
                 package.type == package_type
-                and package.has_label(label)
+                and package.has_labels(labels)
+                and (id is None or package.id == id)
+            ):
+                packages.append(package)
+        return sorted(packages, key=lambda p: p.id)
+
+    def get_user_package(self, package_type, id):
+        packages = self.get_user_packages(package_type, id=id)
+        return packages[0] if packages else None
+
+    def has_user_package(self, package_type, id):
+        return bool(self.get_user_package(package_type, id))
+
+    def has_any_user_package(self, package_type, labels):
+        return bool(self.get_user_packages(package_type, labels=labels))
+
+    def get_packages(self, package_type, os="any", id=None, labels=None):
+        packages = []
+        for package in self.packages:
+            if (
+                package.type == package_type
+                and package.has_labels(labels)
                 and package.is_compatible(os, self.config)
-            )
-        ]
+                and (id is None or package.id == id)
+            ):
+                packages.append(package)
+        return sorted(packages, key=lambda p: p.id)
 
     def get_package(self, package_type, id):
-        for package in self.packages:
-            if package.type == package_type and package.id == id:
-                return package
-        return None
+        packages = self.get_packages(package_type, id=id)
+        return packages[0] if packages else None
 
     def get_formula(self, name):
         for formula in self.formulas:
             if formula.name == name:
                 return formula
-        return None
 
     def find_formulas(self, custom=None, command_name=None):
         for formula in self.formulas:
@@ -99,55 +129,110 @@ class AppMan:
         for pm in self.config.get_compatible_pms(os, package.type):
             if pm in fnames:
                 return self.get_formula(pm)
-        return None
 
-    def _create_packages(self, path, ptype):
-        ptdir = os.path.join(path, ptype)
-        for pdata, ppath in self._load_data_dir(ptdir):
-            pname = self._get_name_from_path(ppath)
-            package = Package(pname, ptype)
-            package.load(pdata)
+    def _create_packages(self, ptype):
+        pname = f"{config.PACKAGES_PKG}.{ptype}"
+        for pfile in self._get_data_resource_files(pname):
+            package = Package(pfile.stem, ptype)
+            package.load(self._load_data_resource(pname, pfile.name))
             yield package
 
-    def _load_data_file(self, path):
-        with open(path, encoding="utf-8") as file:
+    def _load_data_resource(self, package, resource):
+        with resources.open_text(package, resource) as file:
             data = yaml.load(file, Loader=yaml.FullLoader)
-        return data, path
+        return data
 
-    def _load_data_dir(self, path):
-        for root, dirs, files in os.walk(path):
-            for file in files:
-                if file.endswith(".yaml"):
-                    fpath = os.path.join(root, file)
-                    yield self._load_data_file(fpath)
+    def _add_data_resource(self, package, resource, data):
+        content = self._read_data_resource(package, resource)
+        if not content:
+            content = []
+        elif not isinstance(content, list):
+            raise TypeError
 
-    def _load_data_grouped(self, path):
-        for entry in os.scandir(path):
-            if entry.is_file() and entry.path.endswith(".yaml"):
-                data, _ = self._load_data_file(entry.path)
-                ptype = os.path.splitext(entry.name)[0]
-                for d in data:
-                    id = d if isinstance(d, str) else d["id"]
-                    yield {
-                        "id": id,
-                        "data": d,
-                        "ptype": ptype,
-                    }
+        content.append(data)
+        content = sorted(content, key=lambda o: o["id"])
+        self._write_data_resource(package, resource, content)
 
-    def _get_name_from_path(self, path):
-        return os.path.splitext(os.path.basename(path))[0]
+    def _delete_data_resource(self, package, resource, data):
+        content = self._read_data_resource(package, resource)
+        content.remove(data)
+        self._write_data_resource(package, resource, content)
+
+    def _read_data_resource(self, package, resource):
+        fpath = self._get_resource_file(package, resource)
+        if not fpath.exists():
+            return None
+        with open(fpath, encoding="utf-8") as file:
+            return yaml.load(file, Loader=yaml.FullLoader)
+
+    def _write_data_resource(self, package, resource, data):
+        fpath = self._get_resource_file(package, resource)
+        with open(fpath, "w", encoding="utf-8") as file:
+            yaml.dump(data, file)
+
+    def _get_data_resource_files(self, package):
+        path = resources.files(package)
+        yield from path.glob(f"*{config.DEFS_EXT}")
+
+    def _get_grouped_data_resource_files(self, package):
+        path = resources.files(package)
+        for file in path.glob(f"*{config.DEFS_EXT}"):
+            ptype = self._get_package_type(file.stem)
+            data = self._load_data_resource(package, file.name)
+            if not data:
+                continue
+            for d in data:
+                id = d if isinstance(d, str) else d["id"]
+                yield {
+                    "id": id,
+                    "data": d,
+                    "ptype": ptype,
+                }
+
+    def _get_resource_file(self, package, resource):
+        path = resources.files(package)
+        return path.joinpath(f"{resource}{config.DEFS_EXT}")
+
+    def _get_resource_name(self, ptype):
+        return next(pt["pkg"] for pt in config.PACKAGES_TYPES if pt["id"] == ptype)
+
+    def _get_package_type(self, resource):
+        return next(pt["id"] for pt in config.PACKAGES_TYPES if pt["pkg"] == resource)
 
 
-class Package:
+class CommonPackage:
+    def __init__(self, id, ptype):
+        self.id = id
+        self.type = ptype
+        self.labels = []
+
+    def has_labels(self, labels):
+        return not labels or (self.labels and set(labels).issubset(self.labels))
+
+
+class UserPackage(CommonPackage):
+    def __init__(self, id, ptype, labels=None):
+        super().__init__(id, ptype)
+        if labels:
+            self.labels.extend(labels)
+
+    def load(self, obj):
+        if "labels" in obj:
+            self.labels = obj["labels"]
+
+    @property
+    def data(self):
+        return {"id": self.id, "labels": self.labels}
+
+
+class Package(CommonPackage):
     arg_default = "pmid"
 
-    def __init__(self, id, type, format="default"):
-        self.id = id
+    def __init__(self, id, ptype, format="default"):
+        super().__init__(id, ptype)
         self.name = id
-        self.type = type
         self.format = format
         self.description = ""
-        self.labels = []
         self.args = {}
         self.os = []
 
@@ -186,9 +271,6 @@ class Package:
             f"Command '{commandtype}' not found in formula '{formula.name}'"
         )
 
-    def has_label(self, label):
-        return not label or (self.labels and label in self.labels)
-
     def is_compatible(self, os, config):
         if self.format == "simple":
             return True
